@@ -54,6 +54,8 @@ struct _BattfinkClientPriv {
         GSList        *windows;
         BattfinkApm   *fink_apm;
         EggTrayIcon   *tray_icon;
+
+        gint           battery_warning_timeout;
 };
 
 GConfClient *
@@ -130,12 +132,44 @@ battfink_battery_icon_changed_callback (GConfClient *client, guint id, GConfEntr
 }
 
 static void
-battfink_client_invoke_preferences (BattfinkApm *fink_apm)
+battfink_battery_label_changed_callback (GConfClient *client, guint id, GConfEntry *entry, BattfinkClient *fink)
+{
+  gboolean value;
+
+  if (!entry->value || entry->value->type != GCONF_VALUE_BOOL)
+    return;
+
+  value = gconf_value_get_bool (entry->value);
+
+  if (value)
+    BATTFINK_APM_GET_CLASS (fink->priv->fink_apm)->enable_label (fink->priv->fink_apm, fink->priv->tray_icon);
+  else
+    BATTFINK_APM_GET_CLASS (fink->priv->fink_apm)->disable_label (fink->priv->fink_apm, fink->priv->tray_icon);
+}
+
+static void battfink_client_init_dialog (GladeXML *glade_xml, BattfinkClient *fink_client);
+static GladeXML * battfink_client_load_xml (void);
+
+static void
+battfink_client_invoke_preferences (BattfinkApm *fink_apm, gpointer user_data)
 {
   GtkWindow *window;
+  BattfinkClient *fink_client;
+  GladeXML *glade_xml;
+  GtkWidget *dialog;
 
   window = battfink_client_get_active_window (BATTFINK_CLIENT (fink_apm->parent));
-  gtk_window_present (window);
+  if (window) {
+    gtk_window_present (window);
+    return;
+  }
+
+  fink_client = BATTFINK_CLIENT (fink_client);
+  glade_xml = battfink_client_load_xml ();
+  dialog = glade_xml_get_widget (glade_xml, "battfink_dialog");
+  battfink_client_init_dialog (glade_xml, fink_client);
+
+  fink_client->priv->windows = g_slist_append (fink_client->priv->windows, GTK_WINDOW (dialog));
 }
 
 static void
@@ -150,12 +184,25 @@ battfink_client_disable_icon (BattfinkApm *fink_apm)
                          FALSE, NULL);
 }
 
+static void
+battfink_client_disable_label (BattfinkApm *fink_apm)
+{
+  GConfClient *client;
+
+  client = battfink_client_get_gconf_client ();
+
+  gconf_client_set_bool (client,
+                         BATTFINK_GCONF_KEY "show_label",
+                         FALSE, NULL);
+}
+
 static void
 battfink_client_init (BattfinkClient *fink)
 {
   GConfClient *client;
   gboolean battery_value = FALSE;
-
+  gboolean label_value = FALSE;
+  
   fink->priv = g_new0 (BattfinkClientPriv, 1);
   fink->priv->fink_apm = g_object_new (BATTFINK_TYPE_APM, NULL);
 
@@ -166,9 +213,11 @@ battfink_client_init (BattfinkClient *fink)
 
   fink->priv->tray_icon = egg_tray_icon_new ("Battfink");
   g_signal_connect (G_OBJECT (fink->priv->fink_apm), "preferences",
-                    G_CALLBACK (battfink_client_invoke_preferences), NULL);
+                    G_CALLBACK (battfink_client_invoke_preferences), fink);
   g_signal_connect (G_OBJECT (fink->priv->fink_apm), "disable",
                     G_CALLBACK (battfink_client_disable_icon), NULL);
+  g_signal_connect (G_OBJECT (fink->priv->fink_apm), "disable_label",
+                    G_CALLBACK (battfink_client_disable_label), NULL);
 
   gtk_container_add (GTK_CONTAINER (fink->priv->tray_icon), GTK_WIDGET (fink->priv->fink_apm));
 
@@ -182,6 +231,15 @@ battfink_client_init (BattfinkClient *fink)
                            fink, NULL, NULL);
   if (battery_value)
     BATTFINK_APM_GET_CLASS (fink->priv->fink_apm)->enable (fink->priv->fink_apm, fink->priv->tray_icon);
+  
+  label_value = gconf_client_get_bool (client,
+                                       BATTFINK_GCONF_KEY "show_label", NULL);
+  gconf_client_notify_add (client,
+                           BATTFINK_GCONF_KEY "show_label",
+                           (GConfClientNotifyFunc) battfink_battery_label_changed_callback,
+                           fink, NULL, NULL);
+  if (label_value)
+    BATTFINK_APM_GET_CLASS (fink->priv->fink_apm)->enable_label (fink->priv->fink_apm, fink->priv->tray_icon);
 #endif
 }
 
@@ -194,6 +252,7 @@ battfink_client_finalize (GObject *object)
 
   fink = BATTFINK_CLIENT (object);
 
+  g_source_remove (fink->priv->battery_warning_timeout);
   g_return_if_fail (BATTFINK_IS_CLIENT (fink));
   g_return_if_fail (fink->priv != NULL);
 
@@ -212,7 +271,7 @@ battfink_client_new (void)
   return fink;
 }
 
-GladeXML *
+static GladeXML *
 battfink_client_load_xml (void)
 {
   GladeXML *xml = NULL;
@@ -273,6 +332,24 @@ battfink_dialog_destroy_callback (GtkDialog *dialog, gpointer user_data)
 
 static void
 battfink_battery_toggle_callback (GtkToggleButton *toggle, gpointer user_data)
+{
+  GConfClient *client;
+  GladeXML *glade_xml = GLADE_XML (user_data);
+  gboolean state;
+  GtkWidget *label_toggle = glade_xml_get_widget (glade_xml, "battfink_label_checkbox");
+
+  client = battfink_client_get_gconf_client ();
+
+  state = gtk_toggle_button_get_active (toggle);
+  gtk_widget_set_sensitive (label_toggle, state);
+
+  gconf_client_set_bool (client,
+                         BATTFINK_GCONF_KEY "show_battery_status",
+                         state, NULL);
+}
+
+static void
+battfink_battery_label_toggle_callback (GtkToggleButton *toggle, gpointer user_data)
 {
   GConfClient *client;
   gboolean state;
@@ -282,7 +359,7 @@ battfink_battery_toggle_callback (GtkToggleButton *toggle, gpointer user_data)
   state = gtk_toggle_button_get_active (toggle);
 
   gconf_client_set_bool (client,
-                         BATTFINK_GCONF_KEY "show_battery_status",
+                         BATTFINK_GCONF_KEY "show_label",
                          state, NULL);
 }
 
@@ -336,21 +413,17 @@ battfink_client_check_dpms (gpointer user_data)
   GError *error = NULL;
 
   gint standby, suspend, shutdown;
+  static gint standby_old, suspend_old, shutdown_old;
+
   gboolean power_state;
   gboolean value;
 
-  toggle = glade_xml_get_widget (glade_xml, "battfink_display_checkbox");
-  table = glade_xml_get_widget (glade_xml, "battfink_display_table");
-
-  standby_spin = glade_xml_get_widget (glade_xml, "battfink_display_standby_spinbutton");
-  suspend_spin = glade_xml_get_widget (glade_xml, "battfink_display_suspend_spinbutton");
-  shutdown_spin = glade_xml_get_widget (glade_xml, "battfink_display_shutdown_spinbutton");
-  
   power_state = battfink_check_dpms_enabled (&error);
 
   if (error) {
     g_warning ("Failed to check DPMS : %s", error->message);
     g_error_free (error);
+    return TRUE;
   }
   error = NULL;
 
@@ -359,6 +432,26 @@ battfink_client_check_dpms (gpointer user_data)
   if (error) {
     g_warning ("Failed to get DPMS timeouts: %s", error->message);
     g_error_free (error);
+    return TRUE;
+  }
+
+  if (standby == standby_old && suspend == suspend_old && shutdown == shutdown_old)
+    return TRUE;
+ 
+  standby_old = standby;
+  suspend_old = suspend;
+  shutdown_old = shutdown;
+
+  toggle = glade_xml_get_widget (glade_xml, "battfink_display_checkbox");
+  table = glade_xml_get_widget (glade_xml, "battfink_display_table");
+
+  standby_spin = glade_xml_get_widget (glade_xml, "battfink_display_standby_spinbutton");
+  suspend_spin = glade_xml_get_widget (glade_xml, "battfink_display_suspend_spinbutton");
+  shutdown_spin = glade_xml_get_widget (glade_xml, "battfink_display_shutdown_spinbutton");
+  
+  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (toggle)) != power_state) {
+    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle), power_state);
+    gtk_widget_set_sensitive (table, power_state);
   }
 
   /* Let's fill in the values as minutes */
@@ -366,11 +459,6 @@ battfink_client_check_dpms (gpointer user_data)
   gtk_spin_button_set_value (GTK_SPIN_BUTTON (suspend_spin), (gdouble) suspend / 60.0);
   gtk_spin_button_set_value (GTK_SPIN_BUTTON (shutdown_spin), (gdouble) shutdown / 60.0);
 
-  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (toggle)) != power_state) {
-    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle), power_state);
-    gtk_widget_set_sensitive (table, power_state);
-  }
-
   return TRUE;
 }
 
@@ -378,21 +466,26 @@ static void
 battfink_display_spin_callback (GtkSpinButton *spin, gpointer user_data)
 {
   GladeXML *glade_xml = user_data;
-  GtkWidget *suspend_spin;
-  GtkWidget *standby_spin;
-  GtkWidget *shutdown_spin;
+  static GtkWidget *suspend_spin;
+  static GtkWidget *standby_spin;
+  static GtkWidget *shutdown_spin;
   GError *error = NULL;
   gint suspend, standby, shutdown;
+  gdouble min, max;
 
   standby_spin = glade_xml_get_widget (glade_xml, "battfink_display_standby_spinbutton");
   suspend_spin = glade_xml_get_widget (glade_xml, "battfink_display_suspend_spinbutton");
   shutdown_spin = glade_xml_get_widget (glade_xml, "battfink_display_shutdown_spinbutton");
 
-  standby = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (standby_spin)) * 60;
-  suspend = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (suspend_spin)) * 60;
-  shutdown = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (shutdown_spin)) * 60;
+  standby = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (standby_spin));
+  suspend = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (suspend_spin));
+  shutdown = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (shutdown_spin));
 
-  battfink_set_dpms_timeouts (standby, suspend, shutdown, NULL);
+  gtk_spin_button_set_range (GTK_SPIN_BUTTON (standby_spin), 5.0, (gdouble) suspend);
+  gtk_spin_button_set_range (GTK_SPIN_BUTTON (suspend_spin), (gdouble) standby, (gdouble) shutdown);
+  gtk_spin_button_set_range (GTK_SPIN_BUTTON (shutdown_spin), (gdouble) suspend, 300.0);
+  
+  battfink_set_dpms_timeouts (standby * 60, suspend * 60, shutdown * 60, NULL);
 
   if (error) {
     g_warning ("Failed to set DPMS timeouts : %s", error->message);
@@ -401,7 +494,49 @@ battfink_display_spin_callback (GtkSpinButton *spin, gpointer user_data)
 }
 
 static void
-battfink_client_init_dialog (GladeXML *glade_xml)
+battfink_client_warning_response (GtkWidget *widget, gint response, gpointer user_data)
+{
+  switch (response) {
+    default:
+      gtk_widget_destroy (widget);
+  }
+}
+
+static gboolean
+battfink_client_check_battery (gpointer user_data)
+{
+  BattfinkClient *fink;
+  GtkWidget *widget;
+  static gboolean warning = FALSE;
+  gint time_left;
+ 
+  fink = BATTFINK_CLIENT (user_data);
+  time_left = BATTFINK_APM_GET_CLASS (fink->priv->fink_apm)->time_left (fink->priv->fink_apm);
+
+  if (time_left > 0) {
+    if (!warning && time_left < 10) {
+      /* We warn the user that there is not much time left on the battery */
+      widget = gtk_message_dialog_new (fink->priv->active_window,
+                                       GTK_DIALOG_DESTROY_WITH_PARENT,
+                                       GTK_MESSAGE_WARNING,
+                                       GTK_BUTTONS_OK,
+                                       _("Your battery is critically low.\n"
+                                         "You should switch to AC power immediately to avoid losing your work"));
+      g_signal_connect (widget, "response",
+                        G_CALLBACK (battfink_client_warning_response), NULL);
+      gtk_widget_show (widget);
+      warning = TRUE;
+    }
+  } else {
+    warning = FALSE;
+  }
+
+  return TRUE;
+}
+
+
+static void
+battfink_client_init_dialog (GladeXML *glade_xml, BattfinkClient *fink_client)
 {
   GConfClient *client;
 
@@ -409,12 +544,13 @@ battfink_client_init_dialog (GladeXML *glade_xml)
   GtkWidget *display_toggle;
   GtkWidget *battery_vbox;
   GtkWidget *battery_toggle;
+  GtkWidget *label_toggle;
   GtkWidget *display_table;
-  GtkWidget *suspend_spin;
   GtkWidget *standby_spin;
+  GtkWidget *suspend_spin;
   GtkWidget *shutdown_spin;
 
-  gboolean display_value, battery_value = FALSE;
+  gboolean display_value, battery_value = FALSE, label_value = FALSE;
   gint standby, suspend, off;
 
   client = battfink_client_get_gconf_client ();
@@ -423,12 +559,13 @@ battfink_client_init_dialog (GladeXML *glade_xml)
 
   display_toggle = glade_xml_get_widget (glade_xml, "battfink_display_checkbox");
   battery_toggle = glade_xml_get_widget (glade_xml, "battfink_battery_checkbox");
+  label_toggle = glade_xml_get_widget (glade_xml, "battfink_label_checkbox");
   battery_vbox = glade_xml_get_widget (glade_xml, "battfink_battery_vbox");
 
   display_table = glade_xml_get_widget (glade_xml, "battfink_display_table");
 
-  suspend_spin = glade_xml_get_widget (glade_xml, "battfink_display_suspend_spinbutton");
   standby_spin = glade_xml_get_widget (glade_xml, "battfink_display_standby_spinbutton");
+  suspend_spin = glade_xml_get_widget (glade_xml, "battfink_display_suspend_spinbutton");
   shutdown_spin = glade_xml_get_widget (glade_xml, "battfink_display_shutdown_spinbutton");
 
   /* If the value that we get from DPMS is different, then the user must have 
@@ -442,10 +579,21 @@ battfink_client_init_dialog (GladeXML *glade_xml)
 #ifdef HAVE_APM
   battery_value = gconf_client_get_bool (client,
                                          BATTFINK_GCONF_KEY "show_battery_status", NULL);
-  if (battery_value == TRUE)
+  label_value = gconf_client_get_bool (client,
+                                       BATTFINK_GCONF_KEY "show_label", NULL);
+  
+  if (battery_value) {
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (battery_toggle), TRUE);
-  else
+    gtk_widget_set_sensitive (label_toggle, TRUE);
+  } else {
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (battery_toggle), FALSE);
+    gtk_widget_set_sensitive (label_toggle, FALSE);
+  }
+
+  if (label_value)
+    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (label_toggle), TRUE);
+  else
+    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (label_toggle), FALSE);
 
   gconf_client_notify_add (client,
                            BATTFINK_GCONF_KEY "show_battery_status",
@@ -453,7 +601,17 @@ battfink_client_init_dialog (GladeXML *glade_xml)
                            battery_toggle, NULL, NULL);
 
   g_signal_connect (G_OBJECT (battery_toggle), "toggled",
-                    G_CALLBACK (battfink_battery_toggle_callback), NULL);
+                    G_CALLBACK (battfink_battery_toggle_callback), glade_xml);
+  
+  gconf_client_notify_add (client,
+                           BATTFINK_GCONF_KEY "show_label",
+                           (GConfClientNotifyFunc) battfink_battery_changed_callback,
+                           label_toggle, NULL, NULL);
+
+  g_signal_connect (G_OBJECT (label_toggle), "toggled",
+                    G_CALLBACK (battfink_battery_label_toggle_callback), NULL);
+
+  fink_client->priv->battery_warning_timeout = g_timeout_add (5000, battfink_client_check_battery, fink_client);
 #else
     /* Hide the battery checkbox stuff */
   gtk_widget_hide (battery_vbox);
@@ -468,10 +626,10 @@ battfink_client_init_dialog (GladeXML *glade_xml)
   g_signal_connect (G_OBJECT (display_toggle), "toggled",
                     G_CALLBACK (battfink_display_toggle_callback), glade_xml);
 
-  g_signal_connect (G_OBJECT (suspend_spin), "value-changed",
-                    G_CALLBACK (battfink_display_spin_callback), glade_xml);
   g_signal_connect (G_OBJECT (standby_spin), "value-changed",
                     G_CALLBACK (battfink_display_spin_callback), glade_xml);
+  g_signal_connect (G_OBJECT (suspend_spin), "value-changed",
+                    G_CALLBACK (battfink_display_spin_callback), glade_xml);
   g_signal_connect (G_OBJECT (shutdown_spin), "value-changed",
                     G_CALLBACK (battfink_display_spin_callback), glade_xml);
 }
@@ -491,7 +649,7 @@ battfink_client_window_new (BattfinkClient *fink_client)
   glade_xml_signal_autoconnect (glade_xml);
   dialog = glade_xml_get_widget (glade_xml, "battfink_dialog");
   gnome_window_icon_set_from_file (GTK_WINDOW (dialog), GNOME_ICONDIR "/battfink.png");
-  battfink_client_init_dialog (glade_xml);
+  battfink_client_init_dialog (glade_xml, fink_client);
 
   fink_client->priv->windows = g_slist_append (fink_client->priv->windows, GTK_WINDOW (dialog));
 
